import logging
from typing import Dict, Any, Optional, Sequence, List, NoReturn

from pint import DimensionalityError

from nexinfosys.command_executors import BasicCommand, subrow_issue_message, CommandExecutionError
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.command_generators import parser_field_parsers, IType
from nexinfosys.command_generators.parser_ast_evaluators import dictionary_from_key_value_list, ast_to_string
from nexinfosys.common.decorators import memoized_method
from nexinfosys.common.helper import strcmp, first, ifnull, UnitConversion, head
from nexinfosys.models.musiasem_concepts import PedigreeMatrix, FactorType, \
    Factor, FactorInProcessorType, Observer, GeographicReference, ProvenanceReference, \
    BibliographicReference, FactorTypesRelationUnidirectionalLinearTransformObservation, Processor
from nexinfosys.models.musiasem_concepts_helper import _create_or_append_quantitative_observation, \
    find_processors_matching_name, find_factor_types_transform_relation


class InterfacesAndQualifiedQuantitiesCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    invert = {"local": "external",
              "environment": "externalenvironment",
              "external": "local",
              "externalenvironment": "environment"}

    @memoized_method(maxsize=None)
    def _get_relative_to_interface(self, p: Processor, relative_to: str):
        try:
            ast = parser_field_parsers.string_to_ast(parser_field_parsers.factor_unit, relative_to)
        except:
            raise CommandExecutionError(
                f"Could not parse the RelativeTo column, value {str(relative_to)}. " + subrow_issue_message(self._subrow))

        relative_to_interface_name = ast_to_string(ast["factor"])

        # rel_unit_name = ast["unparsed_unit"]
        # try:
        #     f_unit = str((ureg(f_unit) / ureg(rel_unit_name)).units)
        # except (UndefinedUnitError, AttributeError) as ex:
        #     raise CommandExecutionError(f"The final unit could not be computed, interface '{f_unit}' / "
        #                                  f"relative_to '{rel_unit_name}': {str(ex)}"+subrow_issue_message(subrow))

        # relative_to_interface = first(p.factors,
        #                               lambda ifc: strcmp(ifc.name, relative_to_interface_name))
        relative_to_interface = p.factors_find(relative_to_interface_name)

        if not relative_to_interface:
            raise CommandExecutionError(f"Interface specified in 'relative_to' column "
                                        f"'{relative_to_interface_name}' has not been found." + subrow_issue_message(
                self._subrow))

        return relative_to_interface

    def _process_row(self, field_values: Dict[str, Any], subrow=None) -> None:
        """
        Process a dictionary representing a row of the Interfaces command. The dictionary can come directly from
        the worksheet or from a dataset.

        :param field_values: dictionary
        """
        # f_processor_name -> p
        # f_interface_type_name -> it
        # f_interface_name -> i
        #
        # IF NOT i AND it AND p => i_name = it.name => get or create "i"
        # IF i AND it AND p => get or create "i", IF "i" exists, i.it MUST BE equal to "it" (IF NOT, error)
        # IF i AND p AND NOT it => get "i" (MUST EXIST)
        f_interface_type_name = field_values.get("interface_type")
        f_interface_name = field_values.get("interface")

        if not f_interface_name:
            if not f_interface_type_name:
                raise CommandExecutionError("At least one of InterfaceType or Interface must be defined"+subrow_issue_message(subrow))

            f_interface_name = f_interface_type_name

        processor = self.find_processor(field_values.get("processor"), subrow)

        # Try to find Interface
        f_orientation = field_values.get("orientation")
        interface_type: Optional[FactorType] = None
        interface: Optional[Factor] = None
        interfaces: Sequence[Factor] = self._glb_idx.get(Factor.partial_key(processor=processor, name=f_interface_name))
        if len(interfaces) == 1:
            interface = interfaces[0]
            logging.debug(f"Interface '{interface.name}' found")
            interface_type = interface.taxon
            if f_interface_type_name and not strcmp(interface_type.name, f_interface_type_name):
                self._add_issue(IType.WARNING, f"The existing Interface '{interface.name}' has the InterfaceType "
                                               f"'{interface_type.name}' which is different from the specified "
                                               f"InterfaceType '{f_interface_type_name}'. Record skipped." +
                                               subrow_issue_message(subrow))
                return
        elif len(interfaces) > 1:
            raise CommandExecutionError(f"Interface '{f_interface_name}' found {str(len(interfaces))} times. "
                                        f"It must be uniquely identified."+subrow_issue_message(subrow))
        elif len(interfaces) == 0:
            # The interface does not exist, create it below
            if not f_orientation:
                raise CommandExecutionError(f"Orientation must be defined for new Interfaces."+subrow_issue_message(subrow))

        # InterfaceType still not found
        if not interface_type:
            interface_type_name = ifnull(f_interface_type_name, f_interface_name)

            # Find FactorType
            # TODO Allow creating a basic FactorType if it is not found?
            interface_types: Sequence[FactorType] = self._glb_idx.get(FactorType.partial_key(interface_type_name))
            if len(interface_types) == 0:
                raise CommandExecutionError(f"InterfaceType '{interface_type_name}' not declared previously"+subrow_issue_message(subrow))
            elif len(interface_types) > 1:
                raise CommandExecutionError(f"InterfaceType '{interface_type_name}' found {str(len(interface_types))} times. "
                                            f"It must be uniquely identified."+subrow_issue_message(subrow))
            else:
                interface_type = interface_types[0]

        # Get attributes default values taken from Interface Type or Processor attributes
        # Rows   : value of (source) "processor.subsystem_type"
        # Columns: value of (target) "interface_type.opposite_processor_type"
        # Cells  : CORRECTED value of "opposite_processor_type"
        # +--------+-------+--------+-------+---------+
        # |        | Local | Env    | Ext   | ExtEnv  |
        # +--------+-------+--------+-------+---------+
        # | Local  | Local | Env    | Ext   | ExtEnv  |
        # | Env    | Local | Env    | Ext   | ExtEnv? |
        # | Ext    | Ext   | ExtEnv | Local | Env     |
        # | ExtEnv | Ext   | ExtEnv | Local | Env?    |
        # +--------+-------+--------+-------+---------+
        if interface_type.opposite_processor_type:
            tmp = interface_type.opposite_processor_type.lower()
            if processor.subsystem_type.lower() in ["local", "environment"]:  # First two rows
                opposite_processor_type = tmp
            else:
                opposite_processor_type = InterfacesAndQualifiedQuantitiesCommand.invert[tmp]
            # TODO in doubt. Maybe these are undefined (values with question mark in the table)
            #  if tmp == "externalenvironment" and processor.subsystem_type.lower() in ["environment", "externalenvironment"]:
            #      pass
        else:
            opposite_processor_type = None

        interface_type_values = {
            "sphere": interface_type.sphere,
            "roegen_type": interface_type.roegen_type,
            "opposite_processor_type": opposite_processor_type
        }

        # Get internal and user-defined attributes in one dictionary
        # Use: value specified in Interfaces ELSE value specified in InterfaceTypes ELSE first value of allowed values
        attributes = {c.name: ifnull(field_values[c.name], ifnull(interface_type_values.get(c.name), head(c.allowed_values)))
                      for c in self._command_fields if c.attribute_of == Factor}

        if not interface:
            # f_list: Sequence[Factor] = self._glb_idx.get(
            #     Factor.partial_key(processor=p, factor_type=ft, orientation=f_orientation))
            #
            # if len(f_list) > 0:
            #     raise CommandExecutionError(f"An interface called '{f_list[0].name}' for Processor '{f_processor_name}'"
            #                                  f" with InterfaceType '{f_interface_type_name}' and orientation "
            #                                  f"'{f_orientation}' already exists"+subrow_issue_message(subrow))

            # Transform text of "interface_attributes" into a dictionary
            interface_attributes = self.transform_text_attributes_into_dictionary(
                field_values.get("interface_attributes"), subrow)
            attributes.update(interface_attributes)

            location = self.get_location(field_values.get("location"), subrow)

            interface = Factor.create_and_append(f_interface_name,
                                                 processor,
                                                 in_processor_type=FactorInProcessorType(
                                                     external=False,
                                                     incoming=False
                                                 ),
                                                 taxon=interface_type,
                                                 geolocation=location,
                                                 tags=None,
                                                 attributes=attributes)
            self._glb_idx.put(interface.key(), interface)
            logging.debug(f"Interface '{interface.name}' created")
        elif not interface.compare_attributes(attributes):
            initial = ', '.join([f"{k}: {interface.get_attribute(k)}" for k in attributes])
            new = ', '.join([f"{k}: {attributes[k]}" for k in attributes])
            name = interface.processor.full_hierarchy_names(self._glb_idx)[0] + ":" + interface.name
            raise CommandExecutionError(f"The same interface '{name}', is being redeclared with different properties. "
                                        f"INITIAL: {initial}; NEW: {new}."+subrow_issue_message(subrow))

        f_unit = field_values.get("unit")
        if not f_unit:
            f_unit = interface_type.unit

        # Unify unit (it must be done before considering RelativeTo -below-, because it adds a transformation to "f_unit")
        f_value = field_values.get("value")
        if f_value is not None and f_unit != interface_type.unit:
            try:
                f_value = UnitConversion.convert(f_value, f_unit, interface_type.unit)
            except DimensionalityError:
                raise CommandExecutionError(
                    f"Dimensions of units in InterfaceType ({interface_type.unit}) and specified ({f_unit}) are not convertible" + subrow_issue_message(
                        subrow))

            f_unit = interface_type.unit

        # Search for a relative_to interface
        f_relative_to = field_values.get("relative_to")
        relative_to_interface: Optional[Factor] = None
        if f_relative_to:
            self._subrow = subrow
            relative_to_interface = self._get_relative_to_interface(interface.processor, f_relative_to)

        if f_value is None and relative_to_interface is not None:
            # Search for a Interface Type Conversion defined in the ScaleChangeMap command
            interface_types_transforms: List[FactorTypesRelationUnidirectionalLinearTransformObservation] = \
                find_factor_types_transform_relation(self._glb_idx, relative_to_interface.taxon, interface.taxon, processor, processor)

            # Overwrite any specified unit, it doesn't make sense without a value, i.e. it cannot be used for conversion
            f_unit = interface.taxon.unit
            if len(interface_types_transforms) == 1:
                f_value = interface_types_transforms[0].scaled_weight
            else:
                interface_types_transforms_message = "an interface type conversion doesn't exist" \
                    if (len(interface_types_transforms) == 0) \
                    else f"{len(interface_types_transforms)} interface type conversions exist"

                f_value = "0"
                self._add_issue(IType.WARNING, f"Field 'value' should be defined for interfaces having a "
                                               f"'RelativeTo' interface, and {interface_types_transforms_message}. "
                                               f"Using value '0'."+subrow_issue_message(subrow))

        # Create quantitative observation
        if f_value is not None:
            f_uncertainty = field_values.get("uncertainty")
            f_assessment = field_values.get("assessment")
            f_pedigree_matrix = field_values.get("pedigree_matrix")
            f_pedigree = field_values.get("pedigree")
            f_time = field_values.get("time")
            f_comments = field_values.get("comments")

            f_source = field_values.get("qq_source")
            # TODO: source is not being used
            source = self.get_source(f_source, subrow)

            # Find Observer
            observer: Optional[Observer] = None
            if f_source:
                observer = self._glb_idx.get_one(Observer.partial_key(f_source))
                if not observer:
                    self._add_issue(IType.WARNING,
                                    f"Observer '{f_source}' has not been found." + subrow_issue_message(subrow))

            # If an observation exists then "time" is mandatory
            if not f_time:
                raise CommandExecutionError(f"Field 'time' needs to be specified for the given observation."+subrow_issue_message(subrow))

            # An interface can have multiple observations if each of them have a different [time, observer] combination
            for observation in interface.quantitative_observations:
                observer_name = observation.observer.name if observation.observer else None
                if strcmp(observation.attributes["time"], f_time) and strcmp(observer_name, f_source):
                    raise CommandExecutionError(
                        f"The interface '{interface.name}' in processor '{interface.processor.name}' already has an "
                        f"observation with time '{f_time}' and source '{f_source}'.")

            self.check_existence_of_pedigree_matrix(f_pedigree_matrix, f_pedigree, subrow)

            # Transform text of "number_attributes" into a dictionary
            number_attributes = self.transform_text_attributes_into_dictionary(field_values.get("number_attributes"),
                                                                               subrow)

            o = _create_or_append_quantitative_observation(interface,
                                                           f_value, f_unit, f_uncertainty, f_assessment,
                                                           f_pedigree, f_pedigree_matrix,
                                                           observer,
                                                           relative_to_interface,
                                                           f_time,
                                                           None,
                                                           f_comments,
                                                           None, number_attributes
                                                           )

            # TODO Register? Disable for now. Observation can be obtained from a pass over all Interfaces
            # glb_idx.put(o.key(), o)

    def transform_text_attributes_into_dictionary(self, text_attributes: str, subrow) -> Dict:
        dictionary_attributes = {}
        if text_attributes:
            try:
                dictionary_attributes = dictionary_from_key_value_list(text_attributes, self._glb_idx)
            except Exception as e:
                raise CommandExecutionError(str(e) + subrow_issue_message(subrow))

        return dictionary_attributes

    def find_processor(self, processor_name, subrow) -> Processor:
        # Find Processor
        # TODO Allow creating a basic Processor if it is not found?
        processors = find_processors_matching_name(processor_name, self._glb_idx)
        # p = find_observable_by_name(processor_name, self._glb_idx)
        # p = self._glb_idx.get(Processor.partial_key(processor_name))
        if len(processors) == 0:
            raise CommandExecutionError(
                "Processor '" + processor_name + "' not declared previously" + subrow_issue_message(subrow))
        elif len(processors) > 1:
            raise CommandExecutionError(
                f"Processor '{processor_name}' declared previously {len(processors)} times" + subrow_issue_message(subrow))

        return processors[0]

    def get_location(self, reference_name, subrow) -> Any:
        reference = None

        if reference_name:
            try:
                # TODO Change to parser for Location (includes references, but also Codes)
                ast = parser_field_parsers.string_to_ast(parser_field_parsers.reference, reference_name)
                ref_id = ast["ref_id"]
                references = self._glb_idx.get(GeographicReference.partial_key(ref_id))
                if len(references) == 1:
                    reference = references[0]
                else:
                    raise CommandExecutionError(f"Reference '{reference_name}' not found" + subrow_issue_message(subrow))
            except:
                reference = reference_name

        return reference

    @memoized_method(maxsize=None)
    def _get_reference(self, ref_name: str):
        try:
            ast = parser_field_parsers.string_to_ast(parser_field_parsers.reference, ref_name)
            ref_id = ast["ref_id"]
            references = self._glb_idx.get(ProvenanceReference.partial_key(ref_id))
            if len(references) == 1:
                reference = references[0]
            else:
                references = self._glb_idx.get(BibliographicReference.partial_key(ref_id))
                if len(references) == 1:
                    reference = references[0]
                else:
                    raise CommandExecutionError(
                        f"Reference '{ref_name}' not found" + subrow_issue_message(self._subrow))
        except:
            # TODO Change when Ref* are implemented
            reference = ref_name + " (not found)"

        return reference

    def get_source(self, reference_name, subrow) -> Any:
        reference = None

        if reference_name:
            self._subrow = subrow
            reference = self._get_reference(reference_name)

        return reference

    def check_existence_of_pedigree_matrix(self, pedigree_matrix: str, pedigree: str, subrow=None) -> NoReturn:
        # Check existence of PedigreeMatrix, if used
        if pedigree_matrix and pedigree:
            pm = self._glb_idx.get(PedigreeMatrix.partial_key(name=pedigree_matrix))
            if len(pm) == 0:
                raise CommandExecutionError("Could not find Pedigree Matrix '" + pedigree_matrix + "'" +
                                            subrow_issue_message(subrow))
            else:
                try:
                    lst = pm[0].get_modes_for_code(pedigree)
                except:
                    raise CommandExecutionError("Could not decode Pedigree '" + pedigree + "' for Pedigree Matrix '"
                                                + pedigree_matrix + "'" + subrow_issue_message(subrow))
        elif pedigree and not pedigree_matrix:
            raise CommandExecutionError("Pedigree specified without accompanying Pedigree Matrix" +
                                        subrow_issue_message(subrow))



import json

from nexinfosys.command_generators import Issue, IType
from nexinfosys.command_generators.parser_ast_evaluators import check_parameter_value
from nexinfosys.command_generators.spreadsheet_command_parsers_v2 import IssueLocation
from nexinfosys.common.helper import create_dictionary, strcmp
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects, State
from nexinfosys.models.musiasem_concepts import Parameter, ProblemStatement, Hierarchy


class ProblemStatementCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        any_error = False
        issues = []
        sheet_name = self._content["command_name"]
        # Obtain global variables in state
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

        scenarios = create_dictionary()

        for r, param in enumerate(self._content["items"]):
            parameter = param["parameter"]
            scenario = param.get("scenario_name")
            p = glb_idx.get(Parameter.partial_key(parameter))
            if len(p) == 0:
                issues.append(Issue(itype=IType.ERROR,
                                    description="The parameter '" + parameter + "' has not been declared previously.",
                                    location=IssueLocation(sheet_name=sheet_name, row=r, column=None)))
                any_error = True
                continue

            p = p[0]
            name = parameter

            value = param.get("parameter_value")

            check_parameter_value(glb_idx, p, value, issues, sheet_name, r)

            description = param.get("description")  # For readability of the workbook. Not used for solving
            if scenario:
                if scenario in scenarios:
                    sp = scenarios[scenario]
                else:
                    sp = create_dictionary()
                    scenarios[scenario] = sp
                sp[name] = value
            else:
                p.current_value = value
                p.default_value = value

        if not any_error:
            solver_parameters = {}  # {p.name: p.current_value for p in glb_idx.get(Parameter.partial_key()) if p.group and strcmp(p.group, "NISSolverParameters")}
            if len(scenarios) == 0:
                scenarios["default"] = create_dictionary()
            ps = ProblemStatement(solver_parameters, scenarios)
            glb_idx.put(ps.key(), ps)

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        if "description" in json_input:
            self._description = json_input["description"]
        return issues



import io
import logging
import mimetypes
import urllib
from typing import List, Optional

import openpyxl
# import koala # An Excel files parser elaborating a graph allowing automatic evaluation

import nexinfosys
from nexinfosys import Issue
from nexinfosys.command_executors import create_command
from nexinfosys.command_generators.parser_spreadsheet_utils import binary_mask_from_worksheet, \
    obtain_rectangular_submatrices
from nexinfosys.common.helper import create_dictionary, first, download_file
from nexinfosys.command_definitions import valid_v2_command_names, commands
from nexinfosys.command_generators.spreadsheet_command_parsers_v2 import parse_command_in_worksheet
from nexinfosys.command_generators import IType


def handle_import_commands(r):

    def load_file(location: str = None):
        """
        Loads a case study file (well, really any file) into a BytesIO object
        :param location: URL of the case study file
        :return: bytes
        """
        f_type = None
        data = None

        if location:
            # Try to load the Dataset from the specified location
            data = download_file(location).getvalue()
            # data = urllib.request.urlopen(location).read()
            # data = io.BytesIO(data)
            # Then, try to read it
            t = mimetypes.guess_type(location, strict=True)
            if t[0] == "text/python":
                f_type = "python"
            elif t[0] == "text/json":
                f_type = "json"
            else:
            # elif t[0] == "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet":
                f_type = "spreadsheet"

        return f_type, data

    workbook = r.get("workbook_name", None)
    worksheets = r.get("worksheets", None)
    sublist2 = [w.strip() for w in worksheets.split(",")] if worksheets else None  # Convert to list of worksheets
    # Read file in memory
    generator_type, file2 = load_file(workbook)
    return generator_type, file2, sublist2


# ############################### #
#  Main function                  #
# ############################### #


def commands_generator_from_ooxml_file(input, state, sublist, stack, ignore_imports) -> nexinfosys.ExecutableCommandIssuesPairType:
    """
    It reads an Office Open XML input
    Yields a sequence of command_executors

    :param input: A bytes input
    :param state: State used to check variables
    :param sublist: List of worksheets to consider. If not defined, ALL worksheets
    :param stack: Stack of nested files. Just pass it...
    :return:
    """
    # Start the Excel reader
    workbook = openpyxl.load_workbook(io.BytesIO(input), data_only=True)

    # Command names (for the "list of commands" command)
    command_names = create_dictionary(data={cmd_name: None for cmd_name in valid_v2_command_names})

    worksheet_to_command = create_dictionary()  # A dictionary to translate a worksheet to an equivalent command
    if sublist:
        # Force reading "ListOfCommands" commands
        for sheet_name in workbook.sheetnames:
            if first(commands, condition=lambda c: c.name == "list_of_commands" and c.regex.search(sheet_name)):
                sublist.append(sheet_name)

    # For each worksheet, get the command type, convert into primitive JSON
    for sheet_number, sheet_name in enumerate(workbook.sheetnames):
        if sublist:
            if sheet_name not in sublist:
                continue

        issues = []
        total_issues: List[Issue] = []
        sheet = workbook[sheet_name]

        c_label: str = None
        c_content = None

        name = sheet.title.strip()

        # Use an equivalent command name
        if name in worksheet_to_command:
            name = worksheet_to_command[name]

        # Extract worksheet matrices
        m = binary_mask_from_worksheet(sheet, False)
        t = obtain_rectangular_submatrices(m, only_remove_empty_bottom=True)
        if len(t) == 0:  # No data
            continue

        t = t[0]  # Take just the first element, a tuple (top, bottom, left, right) representing a rectangular region
        t = (t[0] + 1, t[1] + 1, t[2] + 1, t[3] + 1)  # Indices start at 1

        # v = worksheet_to_numpy_array(sheet)

        # Find which COMMAND to parse, then parse it
        cmd: Optional[nexinfosys.Command] = first(commands, condition=lambda c: c.regex.search(name))

        c_type: str = cmd.name if cmd else None
        if not c_type:
            total_issues.append(Issue(sheet_number, sheet_name, None, IType.WARNING,
                                f"The worksheet name '{sheet_name}' does not have a supported command associated. Skipped."))

        elif c_type == "etl_dataset":
            if sheet.cell(row=t[0], column=t[2]).value:
                t = (1, m.shape[0] + 1, 1, m.shape[1] + 1)
                # Parse to read parameters
                dataset_name = cmd.regex.search(name).group(2)
                issues, c_label, c_content = cmd.parse_function(sheet, t, dataset_name, state)
            else:
                total_issues.append(
                    Issue(sheet_number, sheet_name, c_type, IType.ERROR,
                          f"It seems there are no parameters for the dataset import command at worksheet '{sheet_name}'"))

        elif c_type == "list_of_commands":
            issues, c_label, c_content = parse_command_in_worksheet(sheet, t, None, cmd.name)
            c_type = None
            if IType.ERROR not in [issue.itype for issue in issues]:
                for r in c_content["items"]:
                    worksheet = r.get("worksheet", None)
                    command = r.get("command", None)
                    # Check if valid command
                    if command not in command_names:
                        total_issues.append(
                            Issue(sheet_number, sheet_name, None, IType.ERROR,
                                  "Command '" + command + "' not recognized in List of Commands."))
                    else:
                        worksheet_to_command[worksheet] = command

        elif c_type == "import_commands":
            if not ignore_imports:
                issues, c_label, c_content = parse_command_in_worksheet(sheet, t, None, cmd.name)
                if IType.ERROR not in [issue.itype for issue in issues]:
                    # Declared at this point to avoid circular reference ("parsers_factory" imports "parsers_spreadsheet")
                    from nexinfosys.command_generators.parsers_factory import commands_container_parser_factory
                    # For each line, repeat the import
                    for r in c_content["items"]:
                        generator_type, file2, sublist2 = handle_import_commands(r)
                        yield from commands_container_parser_factory(generator_type, None, file2, state, sublist=sublist2, stack=stack)
                        logging.debug("IMPORT_COMMANDS done")
            else:
                c_type = None
                issues = []

        elif c_type == "mapping":
            groups = cmd.regex.search(name).groups()
            if groups[2] and groups[8]:
                origin = groups[2]
                destination = groups[8]
            elif not groups[2] and not groups[8]:
                origin = None
                destination = None
            else:
                total_issues.append(
                    Issue(sheet_number, sheet_name, c_type, IType.ERROR,
                          f"Either origin or destination are not correctly specified in the sheet name '{sheet_name}'"))

            issues, c_label, c_content = cmd.parse_function(sheet, t, origin, destination)

        elif c_type in ["datasetqry", "datasetdata"]:
            issues, c_label, c_content = cmd.parse_function(sheet, t, sheet_name, state)

        elif c_type == "hierarchy":
            res = cmd.regex.search(name)
            h_type = res.group(2)
            c_label = res.group(3)
            issues, _, c_content = cmd.parse_function(sheet, t, c_label, h_type)

        elif c_type == "data_input":
            group2_name = cmd.regex.search(name).group(2)
            issues, c_label, c_content = cmd.parse_function(sheet, t, group2_name)

        else:
            # GENERIC command parser
            if cmd.parse_function:
                issues, c_label, c_content = cmd.parse_function(sheet, t, sheet_name)
            else:
                issues, c_label, c_content = parse_command_in_worksheet(sheet, t, sheet_name, cmd.name)

        # -------------------------------------------------------------------------------------------------------------
        # Command parsed, now append "issues"
        errors = 0
        if len(issues) > 0:
            for i in issues:
                if isinstance(i, nexinfosys.command_generators.Issue):
                    if i.itype == IType.ERROR:
                        errors += 1
                    issue = i  # Issue(sheet_number, sheet_name, c_type, i.itype, i.description)
                else:
                    if i[0] == 3:
                        errors += 1
                    issue = Issue(sheet_number, sheet_name, c_type, IType(i[0]), i[1])
                total_issues.append(issue)

        if errors == 0:
            try:
                if c_type:
                    cmd, issues = create_command(c_type, c_label, c_content, sheet_name)
                else:
                    cmd = None
                    issues = []
            except Exception as e:
                cmd = None
                issues = [(3, f"Could not create command of type '{c_type}': {e}")]
            if issues:
                for i in issues:
                    if isinstance(i, nexinfosys.command_generators.Issue):
                        issue = i  # Issue(sheet_number, sheet_name, c_type, i.itype, i.description)
                    else:
                        issue = Issue(sheet_number, sheet_name, c_type, IType(i[0]), i[1])

                    total_issues.append(issue)

        else:
            logging.debug(issues)  # Convenient for debugging purposes
            cmd = None  # cmd, _ = create_command(c_type, c_label, {}, sh_name)

        yield cmd, total_issues
    # yield from []  # Empty generator


# def get_codes_all_statistical_datasets(source, dataset_manager):
#     """
#     Obtain a list of datasets available from a source
#     If no source is specified, all the sources are queried
#     For each dataset, the source, the name, the periods available, an example command and a description are obtained
#
#     :param source:
#     :param dataset_manager: It is a DataSourceManager
#     :return: A Dataframe with the list of datasets
#     """
#     lst2 = []
#     # TODO Probably "get_datasets" will not work as expected. It returns a tuple (Source, list of datasets)
#     for r, k in enumerate(dataset_manager.get_datasets(source)):
#         if len(k) == 4:
#             src = k[3]
#         else:
#             src = ""
#         lst2.append((k[0], k[1], k[2], src))
#     return pd.DataFrame(data=lst2, columns=["Dataset ID", "Description", "URN", "Data Source"])




######################################
#  LIST OF FIELDS FOR THE COMMANDS   #
######################################
import logging
from typing import Dict, List, Type

from nexinfosys import CommandField
from nexinfosys.command_definitions import valid_v2_command_names, commands
from nexinfosys.command_generators.parser_field_parsers import simple_ident, unquoted_string, alphanums_string, \
    hierarchy_expression_v2, key_value_list, key_value, expression_with_parameters, \
    time_expression, indicator_expression, code_string, simple_h_name, domain_definition, unit_name, url_parser, \
    processor_names, value, list_simple_ident, reference, processor_name, processors_selector_expression, \
    interfaces_list_expression, attributes_list_expression, indicators_list_expression, number_interval, pair_numbers, \
    external_ds_name, level_name, expression_with_parameters_or_list_simple_ident, signed_float
from nexinfosys.common.constants import SubsystemType, Scope
from nexinfosys.common.helper import first, class_full_name
from nexinfosys.model_services import IExecutableCommand
from nexinfosys.models.musiasem_concepts import Processor, Factor, RelationClassType, FactorType

data_types = ["Number", "Boolean", "URL", "UUID", "Datetime", "String", "UnitName", "Code", "Geo"]
concept_types = ["Dimension", "Measure", "Attribute", "Dataset"]
parameter_types = ["Number", "Code", "Boolean", "String"]
element_types = ["Parameter", "Processor", "InterfaceType", "Interface"]
spheres = ["Biosphere", "Technosphere"]
roegen_types = ["Flow", "Fund"]
orientations = ["Input", "Output"]
yes_no = ["Yes", "No"]  # Default "Yes"
no_yes = ["No", "Yes"]  # Default "No"
functional_or_structural = ["Functional", "Structural", "Notional"]
instance_or_archetype = ["Yes", "No", "Instance",
                         "Archetype"]  # Yes/No -> Instead of "instance/archetype", now the field is "Accounted"
copy_interfaces_mode = ["No", "FromParent", "FromChildren", "Bidirectional"]
source_cardinalities = ["One", "Zero", "ZeroOrOne", "ZeroOrMore", "OneOrMore"]
target_cardinalities = source_cardinalities
processor_scaling_types = ["CloneAndScale", "Scale", "Clone", "CloneScaled"]
agent_types = ["Person", "Software", "Organization"]
geographic_resource_types = ["dataset"]
geographic_topic_categories = ["Farming", "Biota", "Boundaries", "Climatology", "Meteorology", "Atmosphere", "Economy",
                               "Elevation", "Environment", "GeoscientificInformation", "Health", "Imagery", "BaseMaps",
                               "EarthCover", "Intelligence", "Military", "InlandWaters", "Location", "Oceans",
                               "Planning", "Cadastre", "Society", "Structure", "Transportation", "Utilities",
                               "Communication"]
bib_entry_types = ["article", "book", "booklet", "conference", "inbook", "incollection", "inproceedings",
                   "manual", "mastersthesis", "misc", "phdtesis", "proceedings", "techreport", "unpublished"]
bib_months = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
benchmark_groups = ["Feasibility", "Viability", "Desirability"]
aggregators_list = ["Sum", "Avg", "Count", "SumNA", "CountAv", "AvgNA", "PctNA"]

attributeRegex = "@.+"

# Version 2 only
command_fields: Dict[str, List[CommandField]] = {

    "cat_hierarchies": [
        CommandField(allowed_names=["Source"], name="source", parser=simple_h_name),
        CommandField(allowed_names=["HierarchyGroup"], name="hierarchy_group", parser=simple_ident),
        CommandField(allowed_names=["Hierarchy", "HierarchyName"], name="hierarchy_name", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=["Level", "LevelCode"], name="level", parser=alphanums_string),
        CommandField(allowed_names=["ReferredHierarchy"], name="referred_hierarchy", parser=simple_h_name),
        CommandField(allowed_names=["Code"], name="code", parser=code_string),
        # NOTE: Removed because parent code must be already a member of the hierarchy being defined
        # CommandField(allowed_names=["ReferredHierarchyParent"], name="referred_hierarchy_parent", parser=simple_ident),
        CommandField(allowed_names=["ParentCode"], name="parent_code", parser=code_string),
        CommandField(allowed_names=["Label"], name="label", parser=unquoted_string),
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string),
        CommandField(allowed_names=["Expression", "Formula"], name="expression", parser=hierarchy_expression_v2),
        CommandField(allowed_names=["GeolocationRef"], name="geolocation_ref", parser=reference),
        CommandField(allowed_names=["GeolocationCode"], name="geolocation_code", parser=code_string),
        CommandField(allowed_names=[attributeRegex], name="attributes", many_appearances=True, parser=value),
        CommandField(allowed_names=["Attributes"], name="attributes", parser=key_value_list)
    ],

    "cat_hier_mapping": [
        CommandField(allowed_names=["OriginDataset"], name="source_dataset", parser=external_ds_name),
        CommandField(allowed_names=["OriginHierarchy"], name="source_hierarchy", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["OriginCode"], name="source_code", mandatory=True, parser=code_string),
        CommandField(allowed_names=["DestinationHierarchy"], name="destination_hierarchy", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=["DestinationCode"], name="destination_code", mandatory=True, parser=code_string),
        CommandField(allowed_names=["Weight"], name="weight", mandatory=False, default_value="1",
                     parser=expression_with_parameters),
    ],

    "attribute_types": [
        CommandField(allowed_names=["AttributeType", "AttributeTypeName"], name="attribute_type_name", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=["Type"], name="data_type", mandatory=True, allowed_values=data_types,
                     parser=simple_ident),
        CommandField(allowed_names=["ElementTypes"], name="element_types", default_value=element_types[0],
                     allowed_values=element_types, parser=list_simple_ident),
        CommandField(allowed_names=["Domain"], name="domain", parser=domain_definition)
        # "domain_definition" for Category and NUmber. Boolean is only True or False. Other data types cannot be easily constrained (URL, UUID, Datetime, Geo, String)
    ],

    "datasetdef": [
        CommandField(allowed_names=["Dataset", "DatasetName"], name="dataset_name", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=["DatasetDataLocation"], name="dataset_data_location", parser=url_parser),
        CommandField(allowed_names=["ConceptType"], name="concept_type", mandatory=True, allowed_values=concept_types,
                     parser=simple_ident),
        CommandField(allowed_names=["Concept", "ConceptName"], name="concept_name", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=["ConceptDataType", "DataType"], name="concept_data_type", mandatory=True,
                     allowed_values=data_types, parser=simple_ident),
        CommandField(allowed_names=["ConceptDomain", "Domain"], name="concept_domain", parser=domain_definition),
        CommandField(allowed_names=["ConceptDescription", "Description"], name="concept_description",
                     parser=unquoted_string),
        CommandField(allowed_names=[attributeRegex], name="attributes", many_appearances=True, parser=value),
        CommandField(allowed_names=["ConceptAttributes", "Attributes"], name="attributes", parser=key_value_list)
    ],

    "attribute_sets": [
        CommandField(allowed_names=["AttributeSetName"], name="attribute_set_name", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=[attributeRegex], name="attributes", many_appearances=True, parser=value),
        CommandField(allowed_names=["Attributes"], name="attributes", parser=key_value_list)
    ],

    "parameters": [
        CommandField(allowed_names=["Parameter", "ParameterName"], name="name", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["Type"], name="type", mandatory=True, allowed_values=parameter_types,
                     parser=simple_ident),
        CommandField(allowed_names=["Domain"], name="domain", parser=domain_definition),
        CommandField(allowed_names=["Value"], name="value", parser=expression_with_parameters),
        CommandField(allowed_names=["Group"], name="group", parser=simple_ident),
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string),
        CommandField(allowed_names=[attributeRegex], name="attributes", many_appearances=True, parser=value),
        CommandField(allowed_names=["Attributes"], name="attributes", parser=key_value_list)
    ],

    "interface_types": [
        CommandField(allowed_names=["InterfaceTypeHierarchy"], name="interface_type_hierarchy", parser=simple_ident),
        CommandField(allowed_names=["InterfaceType"], name="interface_type", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["ParentInterfaceType"], name="parent_interface_type", parser=simple_ident),
        CommandField(allowed_names=["Sphere"], name="sphere", mandatory=True, allowed_values=spheres,
                     parser=simple_ident),
        CommandField(allowed_names=["RoegenType"], name="roegen_type", mandatory=True, allowed_values=roegen_types,
                     parser=simple_ident),
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string),
        CommandField(allowed_names=["Source"], name="qq_source", parser=reference),
        # Cristina (in "MuSIASEM Interface List" worksheet)
        CommandField(allowed_names=["Unit"], name="unit", mandatory=True, parser=unit_name),
        CommandField(allowed_names=["OppositeSubsystemType", "OppositeProcessorType"], name="opposite_processor_type",
                     allowed_values=SubsystemType.get_names(), parser=simple_ident),
        CommandField(allowed_names=["Level"], name="level", parser=level_name, attribute_of=FactorType),
        CommandField(allowed_names=["Formula", "Expression"], name="formula", parser=unquoted_string),
        CommandField(allowed_names=[attributeRegex], name="attributes", many_appearances=True, parser=value),
        CommandField(allowed_names=["Attributes"], name="attributes", parser=key_value_list)
    ],

    "processors": [
        CommandField(allowed_names=["ProcessorGroup"], name="processor_group", parser=simple_ident),
        CommandField(allowed_names=["Processor"], name="processor", mandatory=True, parser=processor_name),
        CommandField(allowed_names=["ParentProcessor"], name="parent_processor", parser=processor_name),
        # CommandField(allowed_names=["CopyInterfaces"], name="copy_interfaces_mode",
        #              default_value=copy_interfaces_mode[0], allowed_values=copy_interfaces_mode, parser=simple_ident),
        # CommandField(allowed_names=["CloneProcessor"], name="clone_processor", parser=simple_ident),
        CommandField(allowed_names=["SubsystemType", "ProcessorContextType", "ProcessorType"], name="subsystem_type",
                     default_value=SubsystemType.get_names()[0], allowed_values=SubsystemType.get_names(),
                     parser=simple_ident,
                     attribute_of=Processor),
        CommandField(allowed_names=["System"], name="processor_system", default_value="default",
                     parser=simple_ident, attribute_of=Processor),
        CommandField(allowed_names=["FunctionalOrStructural"], name="functional_or_structural",
                     default_value=functional_or_structural[0], allowed_values=functional_or_structural,
                     parser=simple_ident, attribute_of=Processor),
        CommandField(allowed_names=["Accounted", "InstanceOrArchetype"], name="instance_or_archetype",
                     default_value=instance_or_archetype[0], allowed_values=instance_or_archetype, parser=simple_ident,
                     attribute_of=Processor),
        CommandField(allowed_names=["ParentProcessorWeight"], name="parent_processor_weight",
                     parser=expression_with_parameters),
        CommandField(allowed_names=["BehaveAs"], name="behave_as_processor", parser=processor_name),
        CommandField(allowed_names=["Level"], name="level", parser=level_name, attribute_of=Processor),
        CommandField(allowed_names=["Stock"], name="stock", default_value=no_yes[0], allowed_values=no_yes,
                     deprecated=True,
                     parser=simple_ident, attribute_of=Processor),
        # CommandField(allowed_names=["Alias", "SpecificName"], name="alias", parser=simple_ident),
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string),
        CommandField(allowed_names=["GeolocationRef"], name="geolocation_ref", parser=reference),
        CommandField(allowed_names=["GeolocationCode"], name="geolocation_code", parser=code_string),
        CommandField(allowed_names=["GeolocationLatLong"], name="geolocation_latlong", parser=pair_numbers),
        CommandField(allowed_names=[attributeRegex], name="attributes", many_appearances=True, parser=value),
        CommandField(allowed_names=["Attributes"], name="attributes", parser=key_value_list)
    ],

    "interfaces_and_qq": [
        CommandField(allowed_names=["Processor"], name="processor", mandatory=True, parser=processor_name),
        CommandField(allowed_names=["InterfaceType"], name="interface_type", parser=simple_ident),
        CommandField(allowed_names=["Interface"], name="interface", parser=simple_ident),
        CommandField(allowed_names=["Sphere"], name="sphere", allowed_values=spheres, parser=simple_ident,
                     attribute_of=Factor),
        CommandField(allowed_names=["RoegenType"], name="roegen_type", allowed_values=roegen_types, parser=simple_ident,
                     attribute_of=Factor),
        CommandField(allowed_names=["Orientation"], name="orientation", mandatory=True, allowed_values=orientations,
                     parser=simple_ident, attribute_of=Factor),
        CommandField(allowed_names=["OppositeSubsystemType", "OppositeProcessorType"], name="opposite_processor_type",
                     allowed_values=SubsystemType.get_names(), parser=simple_ident, attribute_of=Factor),
        CommandField(allowed_names=["GeolocationRef"], name="geolocation_ref", parser=reference),
        CommandField(allowed_names=["GeolocationCode"], name="geolocation_code", parser=code_string),
        # CommandField(allowed_names=["Alias", "SpecificName"], name="alias", parser=simple_ident),
        CommandField(allowed_names=["I" + attributeRegex], name="interface_attributes", many_appearances=True,
                     parser=value),
        CommandField(allowed_names=["Range"], name="range", parser=number_interval),
        CommandField(allowed_names=["RangeUnit"], name="range_unit", parser=unit_name),
        CommandField(allowed_names=["InterfaceAttributes"], name="interface_attributes", parser=key_value_list),
        # Qualified Quantification
        CommandField(allowed_names=["Value"], name="value", parser=expression_with_parameters),
        CommandField(allowed_names=["Unit"], name="unit", parser=unit_name),
        CommandField(allowed_names=["RelativeTo"], name="relative_to", parser=unquoted_string),
        CommandField(allowed_names=["Uncertainty"], name="uncertainty", parser=unquoted_string),
        CommandField(allowed_names=["Assessment"], name="assessment", parser=unquoted_string),
        # TODO
        # CommandField(allowed_names=["Pedigree"], name="pedigree", parser=pedigree_code),
        # CommandField(allowed_names=["RelativeTo"], name="relative_to", parser=simple_ident_plus_unit_name),
        CommandField(allowed_names=["PedigreeMatrix"], name="pedigree_matrix", parser=reference),
        CommandField(allowed_names=["Pedigree"], name="pedigree", parser=unquoted_string),
        CommandField(allowed_names=["Time"], name="time", parser=time_expression),
        CommandField(allowed_names=["Source"], name="qq_source", parser=reference),
        CommandField(allowed_names=["N" + attributeRegex], name="number_attributes", many_appearances=True,
                     parser=key_value),
        CommandField(allowed_names=["NumberAttributes"], name="number_attributes", parser=key_value_list),
        CommandField(allowed_names=["Comments"], name="comments", parser=unquoted_string)
    ],

    "relationships": [
        CommandField(allowed_names=["OriginProcessors", "OriginProcessor"], name="source_processor", mandatory=True,
                     parser=processor_names),
        CommandField(allowed_names=["OriginInterface"], name="source_interface", parser=simple_ident),
        CommandField(allowed_names=["DestinationProcessors", "DestinationProcessor"], name="target_processor",
                     mandatory=True, parser=processor_names),
        CommandField(allowed_names=["DestinationInterface"], name="target_interface", parser=simple_ident),
        CommandField(allowed_names=["BackInterface"], name="back_interface", parser=simple_ident),
        CommandField(allowed_names=["RelationType"], name="relation_type", mandatory=True,
                     allowed_values=RelationClassType.relationships_command_labels(), parser=unquoted_string),
        CommandField(allowed_names=["Weight"], name="flow_weight", parser=expression_with_parameters),
        CommandField(allowed_names=["ChangeOfTypeScale"], name="change_type_scale", parser=expression_with_parameters),
        CommandField(allowed_names=["OriginCardinality"], name="source_cardinality",
                     default_value=source_cardinalities[0],
                     allowed_values=source_cardinalities, parser=simple_ident),
        CommandField(allowed_names=["DestinationCardinality"], name="target_cardinality",
                     default_value=target_cardinalities[0],
                     allowed_values=target_cardinalities, parser=simple_ident),
        CommandField(allowed_names=["Attributes"], name="attributes", parser=key_value_list)
    ],

    "processor_scalings": [
        CommandField(allowed_names=["InvokingProcessor"], name="invoking_processor", mandatory=True,
                     parser=processor_name),
        CommandField(allowed_names=["RequestedProcessor"], name="requested_processor", mandatory=True,
                     parser=processor_name),
        CommandField(allowed_names=["ScalingType"], name="scaling_type", mandatory=True,
                     allowed_values=processor_scaling_types, parser=simple_ident),
        CommandField(allowed_names=["InvokingInterface"], name="invoking_interface", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=["RequestedInterface"], name="requested_interface", mandatory=True,
                     parser=simple_ident),
        CommandField(allowed_names=["Scale"], name="scale", mandatory=True, parser=expression_with_parameters),
        # BareProcessor fields
        CommandField(allowed_names=["NewProcessorName"], name="new_processor_name", parser=processor_name),
        CommandField(allowed_names=["NewProcessorGroup"], name="processor_group", parser=simple_ident,
                     attribute_of=Processor),
        CommandField(allowed_names=["NewParentProcessor"], name="parent_processor", parser=processor_name),
        CommandField(allowed_names=["NewSubsystemType"], name="subsystem_type",
                     default_value=SubsystemType.get_names()[0], allowed_values=SubsystemType.get_names(),
                     parser=simple_ident,
                     attribute_of=Processor),
        CommandField(allowed_names=["NewProcessorLevel"], name="level", parser=level_name, attribute_of=Processor),
        # DISABLED because this fields apply to processors Cloned and Scaled, so they will always have a parent,
        #          and children inherit the system of the parent.
        # CommandField(allowed_names=["NewSystem"], name="processor_system", default_value="_default_system",
        #              parser=simple_ident, attribute_of=Processor),
        CommandField(allowed_names=["NewDescription"], name="description", parser=unquoted_string,
                     attribute_of=Processor),
        CommandField(allowed_names=["NewGeolocationRef"], name="geolocation_ref", parser=reference,
                     attribute_of=Processor),
        CommandField(allowed_names=["NewGeolocationCode"], name="geolocation_code", parser=code_string,
                     attribute_of=Processor),
        CommandField(allowed_names=[attributeRegex], name="attributes", many_appearances=True, parser=value),
        CommandField(allowed_names=["NewAttributes"], name="attributes", parser=key_value_list)
        # CommandField(allowed_names=["UpscaleParentContext"], name="upscale_parent_context", parser=upscale_context),
        # CommandField(allowed_names=["UpscaleChildContext"], name="upscale_child_context", parser=upscale_context)
    ],

    "scale_conversion_v2": [
        CommandField(allowed_names=["OriginHierarchy"], name="source_hierarchy", parser=simple_ident),
        CommandField(allowed_names=["OriginInterfaceType"], name="source_interface_type", parser=simple_ident),
        CommandField(allowed_names=["DestinationHierarchy"], name="target_hierarchy", parser=simple_ident),
        CommandField(allowed_names=["DestinationInterfaceType"], name="target_interface_type", parser=simple_ident),
        CommandField(allowed_names=["OriginContext"], name="source_context", parser=processor_names),
        CommandField(allowed_names=["DestinationContext"], name="target_context", parser=processor_names),
        CommandField(allowed_names=["Scale"], name="scale", mandatory=True, parser=expression_with_parameters),
        CommandField(allowed_names=["OriginUnit"], name="source_unit", parser=unit_name),
        CommandField(allowed_names=["DestinationUnit"], name="target_unit", parser=unit_name)
    ],

    "import_commands": [
        CommandField(allowed_names=["Workbook", "WorkbookLocation"], name="workbook_name", parser=url_parser),
        CommandField(allowed_names=["Worksheets"], name="worksheets", parser=unquoted_string)
    ],

    "list_of_commands": [
        CommandField(allowed_names=["Worksheet", "WorksheetName"], name="worksheet", mandatory=True,
                     parser=unquoted_string),
        CommandField(allowed_names=["Command"], name="command", mandatory=True, allowed_values=valid_v2_command_names,
                     parser=simple_ident),
        CommandField(allowed_names=["Comment", "Description"], name="comment", parser=unquoted_string)
    ],

    "ref_provenance": [
        # Reduced, from W3C Provenance Recommendation (https://www.w3.org/TR/prov-overview/)
        CommandField(allowed_names=["RefID", "Reference"], name="ref_id", mandatory=True, parser=simple_ident),
        # The reference "RefID" should be mentioned
        CommandField(allowed_names=["ProvenanceFileURL"], name="provenance_file_url", parser=url_parser),
        CommandField(allowed_names=["AgentType"], name="agent_type", mandatory=True, allowed_values=agent_types,
                     parser=simple_ident),
        CommandField(allowed_names=["Agent"], name="agent", mandatory=True, parser=unquoted_string),
        CommandField(allowed_names=["Activities"], name="activities", mandatory=True, parser=unquoted_string),
        CommandField(allowed_names=["Entities"], name="entities", parser=unquoted_string)
    ],

    "ref_geographical": [
        # A subset of fields from INSPIRE regulation for metadata: https://eur-lex.europa.eu/legal-content/EN/TXT/PDF/?uri=CELEX:32008R1205&from=EN
        # Fields useful to elaborate graphical displays. Augment in the future as demanded
        CommandField(allowed_names=["RefID", "Reference"], name="ref_id", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["GeoLayerURL", "DataLocation", "ResourceLocator"], name="data_location",
                     parser=url_parser),
        CommandField(allowed_names=["Title"], name="title", mandatory=True, parser=unquoted_string),
        CommandField(allowed_names=["Description", "Abstract"], name="description", parser=unquoted_string),  # Syntax??
        CommandField(allowed_names=["BoundingBox"], name="bounding_box", parser=unquoted_string),  # Syntax??
        CommandField(allowed_names=["TopicCategory"], name="topic_category",
                     default_value=geographic_topic_categories[0],
                     allowed_values=geographic_topic_categories, parser=unquoted_string),  # Part D.2
        CommandField(allowed_names=["TemporalExtent", "Date"], name="temporal_extent", parser=unquoted_string),
        # Syntax??
        CommandField(allowed_names=["PointOfContact"], name="point_of_contact", parser=unquoted_string),
        CommandField(allowed_names=["Type"], name="type", default_value=geographic_resource_types[0],
                     allowed_values=geographic_resource_types, parser=unquoted_string)  # Part D.1. JUST "Dataset"
    ],

    "ref_bibliographic": [
        # From BibTex. Mandatory fields depending on EntryType, at "https://en.wikipedia.org/wiki/BibTeX" (or search: "Bibtex entry field types")
        CommandField(allowed_names=["RefID", "Reference"], name="ref_id", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["BibFileURL"], name="bib_file_url", parser=url_parser),
        CommandField(allowed_names=["EntryType"], name="entry_type", mandatory=True, allowed_values=bib_entry_types,
                     parser=unquoted_string),
        CommandField(allowed_names=["Address"], name="address", parser=unquoted_string),
        CommandField(allowed_names=["Annote"], name="annote", parser=unquoted_string),
        CommandField(allowed_names=["Author"], name="author",
                     mandatory="entry_type not in ('booklet', 'manual', 'misc', 'proceedings')",
                     parser=unquoted_string),
        CommandField(allowed_names=["BookTitle"], name="booktitle",
                     mandatory="entry_type in ('incollection', 'inproceedings')", parser=unquoted_string),
        CommandField(allowed_names=["Chapter"], name="chapter", mandatory="entry_type in ('inbook')",
                     parser=unquoted_string),
        CommandField(allowed_names=["CrossRef"], name="crossref", parser=unquoted_string),
        CommandField(allowed_names=["Edition"], name="edition", parser=unquoted_string),
        CommandField(allowed_names=["Editor"], name="editor", mandatory="entry_type in ('book', 'inbook')",
                     parser=unquoted_string),
        CommandField(allowed_names=["HowPublished"], name="how_published", parser=unquoted_string),
        CommandField(allowed_names=["Institution"], name="institution", mandatory="entry_type in ('techreport')",
                     parser=unquoted_string),
        CommandField(allowed_names=["Journal"], name="journal", mandatory="entry_type in ('article')",
                     parser=unquoted_string),
        CommandField(allowed_names=["Key"], name="key", parser=unquoted_string),
        CommandField(allowed_names=["Month"], name="month", default_value=bib_months[0], allowed_values=bib_months,
                     parser=simple_ident),
        CommandField(allowed_names=["Note"], name="note", parser=unquoted_string),
        CommandField(allowed_names=["Number"], name="number", parser=unquoted_string),
        CommandField(allowed_names=["Organization"], name="organization", parser=unquoted_string),
        CommandField(allowed_names=["Pages"], name="pages", mandatory="entry_type in ('inbook')",
                     parser=unquoted_string),
        CommandField(allowed_names=["Publisher"], name="publisher",
                     mandatory="entry_type in ('book', 'inbook', 'incollection')", parser=unquoted_string),
        CommandField(allowed_names=["School"], name="school", mandatory="entry_type in ('mastersthesis', 'phdtesis')",
                     parser=unquoted_string),
        CommandField(allowed_names=["Series"], name="series", parser=unquoted_string),
        CommandField(allowed_names=["Title"], name="title", mandatory="entry_type not in ('misc')",
                     parser=unquoted_string),
        CommandField(allowed_names=["Type"], name="type", parser=unquoted_string),
        CommandField(allowed_names=["URL"], name="url", parser=url_parser),
        CommandField(allowed_names=["Volume"], name="volume", mandatory="entry_type in ('article')",
                     parser=unquoted_string),
        CommandField(allowed_names=["Year"], name="year",
                     mandatory="entry_type in ('article', 'book', 'inbook', 'incollection', 'inproceedings', 'mastersthesis', 'phdthesis', 'proceedings', 'techreport')",
                     parser=unquoted_string)
    ],

    # Used only for help elaboration
    "datasetqry": [
        CommandField(allowed_names=["InputDataset"], name="inputdataset", parser=external_ds_name),
        CommandField(allowed_names=["AvailableAtDateTime"], name="availableatdatetime", parser=unquoted_string),
        CommandField(allowed_names=["StartTime"], name="starttime", parser=time_expression),
        CommandField(allowed_names=["EndTime"], name="endtime", parser=time_expression),
        CommandField(allowed_names=["ResultDimensions"], name="resultdimensions", parser=simple_ident),
        CommandField(allowed_names=["ResultMeasures"], name="resultmeasures", parser=simple_ident),
        CommandField(allowed_names=["ResultMeasuresAggregation"], name="resultmeasuresaggregation",
                     default_value=aggregators_list[0], allowed_values=aggregators_list, parser=simple_ident),
        CommandField(allowed_names=["ResultMeasureName"], name="resultmeasurename", parser=simple_ident),
        CommandField(allowed_names=["OutputDataset"], name="outputdataset", parser=simple_ident),
    ],

    # Analysis commands

    "problem_statement": [
        CommandField(allowed_names=["Scenario"], name="scenario_name", parser=simple_ident),
        CommandField(allowed_names=["Parameter"], name="parameter", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["Value"], name="parameter_value", mandatory=True,
                     parser=expression_with_parameters_or_list_simple_ident),  # list_simple_ident
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string)
    ],

    "scalar_indicator_benchmarks": [
        CommandField(allowed_names=["BenchmarkGroup"], name="benchmark_group", default_value=benchmark_groups[0],
                     allowed_values=benchmark_groups, parser=simple_ident),
        CommandField(allowed_names=["Stakeholders"], name="stakeholders", parser=list_simple_ident),
        CommandField(allowed_names=["Benchmark"], name="benchmark", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["Range"], name="range", mandatory=True, parser=number_interval),
        CommandField(allowed_names=["Unit"], name="unit", mandatory=True, parser=unit_name),
        CommandField(allowed_names=["Category"], name="category", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["Label"], name="label", mandatory=True, parser=unquoted_string),
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string)
    ],

    # Modified to consider Cristina's annotations
    "scalar_indicators": [
        CommandField(allowed_names=["IndicatorsGroup"], name="indicators_group", parser=simple_ident),
        # IndicatorType (Indicators)
        CommandField(allowed_names=["Indicator"], name="indicator_name", mandatory=True, parser=simple_ident),
        # IndicatorName (Indicators)
        CommandField(allowed_names=["Local"], name="local", mandatory=True, allowed_values=yes_no, parser=simple_ident),
        CommandField(allowed_names=["Formula", "Expression"], name="formula", mandatory=True,
                     parser=indicator_expression),
        # A call to LCIAMethod function, with parameter IndicatorMethod (Indicators)
        CommandField(allowed_names=["Unit"], name="unit", mandatory=True, parser=unit_name),
        # IndicatorUnit (Indicators)
        CommandField(allowed_names=["AccountNA"], name="account_na", mandatory=False, allowed_values=no_yes,
                     parser=simple_ident),  # IndicatorUnit (Indicators)
        # TODO Disabled: apply the formula to ALL processors (and ignore those where it cannot be evaluated)
        #  CommandField(allowed_names=["Processors"], name="processors_selector", parser=processors_selector_expression)
        CommandField(allowed_names=["Benchmarks", "Benchmark"], name="benchmarks", parser=list_simple_ident),
        CommandField(allowed_names=["UnitLabel"], name="unit_label", mandatory=False, parser=unquoted_string),
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string),
        CommandField(allowed_names=["Reference", "Source"], name="source", mandatory=False, parser=reference),
        # SAME (Indicators)
    ],

    "matrix_indicators": [
        CommandField(allowed_names=["Indicator"], name="indicator_name", mandatory=True, parser=simple_ident),
        CommandField(allowed_names=["Scope"], name="scope", default_value=Scope.get_names()[0],
                     allowed_values=Scope.get_names(), parser=simple_ident),
        CommandField(allowed_names=["Processors"], name="processors_selector", parser=processors_selector_expression),
        CommandField(allowed_names=["Interfaces"], name="interfaces_selector", parser=interfaces_list_expression),
        CommandField(allowed_names=["Indicators"], name="indicators_selector", parser=indicators_list_expression),
        CommandField(allowed_names=["Attributes"], name="attributes_selector", parser=attributes_list_expression),
        CommandField(allowed_names=["Description"], name="description", parser=unquoted_string)
    ],

    # NEW command, implementation of Cristina's suggestions
    "lcia_methods": [
        CommandField(allowed_names=["LCIAMethod"], name="lcia_method", mandatory=True, parser=unquoted_string),
        # IndicatorMethod (Indicators), SAME (LCIAmethod)
        CommandField(allowed_names=["LCIAIndicator"], name="lcia_indicator", mandatory=True, parser=simple_ident),
        # IndicatorName (Indicators), SAME (LCIAmethod)
        CommandField(allowed_names=["LCIAIndicatorUnit"], name="lcia_indicator_unit", mandatory=True,
                     parser=unquoted_string),  # IndicatorName (Indicators), SAME (LCIAmethod)
        CommandField(allowed_names=["LCIAHorizon"], name="lcia_horizon", mandatory=True, parser=simple_ident),
        # IndicatorTemporal (Indicators)
        CommandField(allowed_names=["Interface"], name="interface", mandatory=True, parser=simple_ident),
        # SAME (LCIAmethod)
        CommandField(allowed_names=["Compartment"], name="compartment", mandatory=False, parser=unquoted_string),
        # SAME (LCIAmethod)
        CommandField(allowed_names=["Subcompartment"], name="subcompartment", mandatory=False, parser=unquoted_string),
        # SAME (LCIAmethod)
        CommandField(allowed_names=["InterfaceUnit"], name="interface_unit", mandatory=True, parser=unit_name),
        # Not present, but needed to warrant independence from specification of InterfaceTypes
        CommandField(allowed_names=["LCIACoefficient"], name="lcia_coefficient", mandatory=True,  # SAME (LCIAmethod)
                     parser=signed_float)
    ],

}


def get_command_fields_from_class(execution_class: Type[IExecutableCommand]) -> List[CommandField]:
    execution_class_name: str = class_full_name(execution_class)
    cmd = first(commands, condition=lambda c: c.execution_class_name == execution_class_name)
    if cmd:
        return command_fields.get(cmd.name, [])

    return []


# command_field_names = {}
# for fields in command_fields.values():
#     for field in fields:
#         for name in field.allowed_names:
#             command_field_names[name] = field.name
_command_field_names = {name: f.name for fields in command_fields.values() for f in fields for name in f.allowed_names}



''' Plot FAO data on world bmaps '''
import datetime
import os
import re
import tarfile

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
from mpl_toolkits.basemap import shapefile

from magic_box import source_fao_restful

# Decompress Natural Earth boundaries shapefile
FDATA = os.path.dirname(os.path.abspath(__file__))
FSHP = '%s/ne_110m_admin_0_countries.shp' % FDATA

if not os.path.exists(FSHP):
    with tarfile.open(re.sub('shp$', 'tar.gz', FSHP)) as tar:
        for item in tar:
            tar.extract(item, FDATA)

def get_countries_shapefiledata():
    ''' Get country meta-data from Natural Earth shapefile
    see http://www.naturalearthdata.com/downloads/110m-cultural-vectors/

    Returns
    -----------
    data : pandas.core.series.Series
        Country Data

    Example
    -----------
    >>> df = faobmap.get_countries_shapefiledata()
    >>> df.shape
    (255, 65)
    '''
    shf = shapefile.Reader(re.sub('\\.shp', '', FSHP))

    cols = [f[0] for f in shf.fields[1:]]

    data = []
    for shprec in shf.shapeRecords():
        rec = {k:v for k, v in zip(cols, shprec.record)}
        data.append(rec)

    data = pd.DataFrame(data)

    # Fix missing ISO 3 country code
    idx = data['iso_a3'] == '-99'
    data.loc[idx, 'iso_a3'] = data.loc[idx, 'wb_a3']

    return data


def plot(bmap, data, cat, \
        country_field='country', \
        value_field='value', \
        cbmap=plt.get_cmap('Blues'), \
        ndigits=0):
    ''' Draw polygon shapefile.
        Arguments sent to PatchCollection constructor

    Parameters
    -----------
    bmap : mpl_toolkits.basebmap.Basebmap
        Axe to draw data on
    data : pandas.core.series.Series
        Data to be plotted
    cat : list
        Categories
    country_field : str
        Name of field in data containing ISO3 country codes
    value_field : str
        Name of field in data containing data to be plotted
    cbmap : matplotlib.colors.Colorbmap
        Color bmap to use for each category
    ndigits : int
        Number of digits to be displayed in legend

    Example
    -----------
    >>> import numpy as np
    >>> from hygis import oz
    >>> import matplotlib.pyplot as plt
    >>> nval = 200
    '''

    # Check inputs
    nbc = data.groupby(country_field).apply(len)
    if len(nbc[nbc > 1]) > 0:
        raise ValueError('Some countries are reported ' \
                'more than once in dataset')

    # Add country polygons to basebmap
    shapefilepath = os.path.basename(FSHP)
    bmap.readshapefile(re.sub('\\.shp$', '', FSHP), shapefilepath, \
            drawbounds=False)

    # Correct cat to account for min/max
    if cat[0] > data[value_field].min():
        cat = [data[value_field].min()] + cat

    if cat[-1] < data[value_field].max():
        cat = cat + [data[value_field].max()]

    # Build metadata dataframe
    cdata = get_countries_shapefiledata()
    cdata = cdata.rename(columns={'iso_a3':country_field})
    cdata = pd.merge(cdata, data, how='left', on=country_field)
    cdata['icat'] = pd.cut(cdata[value_field], cat, labels=False)
    cdata['icat'] = cdata['icat'].fillna(-1)

    # Initialise
    ncat = len(cat)-1
    colors = {i:cbmap(int(float(i)/ncat*cbmap.N)) for i in range(ncat)}
    colors[-1] = 'lightgrey'
    patches = {-1:[]}

    for i in range(ncat):
        patches[i] = []

    # Build patches
    for info, shape in zip(getattr(bmap, '%s_info' % shapefilepath), \
          getattr(bmap, shapefilepath)):

        iso = info['wb_a3']
        if iso == '-99':
            iso = info['iso_a3']

        idx = cdata[country_field] == iso
        if np.sum(idx) >= 1:
            icat = cdata.loc[idx, 'icat'].values[0]
            patches[icat].append(Polygon(np.array(shape), True))


    # plot patches with appropriate color
    for i in patches.keys():
        if len(patches[i]) > 0:
            # Add patch
            pac = PatchCollection(patches[i], \
                    edgecolor='none', facecolor=colors[i])
            bmap.ax.add_collection(pac)

            # Add legend entry
            if i == -1:
                label = 'No data'
            else:
                ndtxt = str(ndigits)
                label = ('%0.' + ndtxt + 'f to %0.' \
                            + ndtxt + 'f') % (cat[i], cat[i+1])
            bmap.ax.plot([], [], lw=10, color=colors[i], \
                    label=label)


def mapfooter(fig, database, dataset, field):
    ''' Add footer to a figure (code from wafari) '''

    now = datetime.datetime.now()
    label = 'source: FAO (%s, %s, %s)\nurl: %s\ngenerated: %s' % ( \
        database, dataset, field, \
        source_fao_restful.__fao_url__, \
        now.strftime('%d/%m/%Y %H:%M'))

    # Add label
    props = {
        'boxstyle':'round',
        'edgecolor':'none',
        'facecolor':'white'
    }

    fig.text(0.77, 0.03, label, \
            bbox=props, \
            color='#595959', \
            ha='left', fontsize=9)




from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import IssuesLabelContentTripleType, AreaTupleType
from nexinfosys.command_generators.parser_spreadsheet_utils import obtain_rectangular_submatrices, \
    binary_mask_from_worksheet


def parse_upscale_command(sh: Worksheet, area: AreaTupleType, name: str = None) -> IssuesLabelContentTripleType:
    """
    Analyze the input area
    Obtain the numerical part
    Obtain the tags identifying parent and child processors. Some tags may be used for both

    Read
    Most "parse" methods are mostly syntactic (as opposed to semantic). They do not check existence of names.
    But in this case, the valid field names are fixed beforehand, so they are checked at this time.
    Some of the fields will be controlled also, according to some

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :return: list of issues (issue_type, message), command label, command content
    """

    def get_subrow(r, c1, c2):
        lst = []
        # To deal with combined cell ranges, store "previous" value, and if "" is found, assume it is a merged cell
        previous = None
        for c in range(c1, c2):
            v = sh.cell(row=r, column=c).value
            if not v:
                if previous:
                    lst.append(previous)
                else:
                    lst.append("")
            else:
                previous = v
                lst.append(v)

        return lst

    def get_subcolumn(c, r1, r2):
        lst = []
        # To deal with combined cell ranges, store "previous" value, and if "" is found, assume it is a merged cell
        # !!! This may not be correct at all times: when a cell is intentionally left blank
        # To solve this, use "sh.merged_cell_ranges" to check if the current cell (r, c) is inside a range
        previous = None
        for r in range(r1, r2):
            v = sh.cell(row=r, column=c).value
            if not v:
                if previous:
                    lst.append(previous)
                else:
                    lst.append("")
            else:
                previous = v
                lst.append(v)
        return lst

    some_error = False
    issues = []

    # Detect the matrix defining scales
    m = binary_mask_from_worksheet(sh, True)  # "True" is to focus on cells containing numbers
    # Locate the matrix with numbers. Assume this defines the labels to consider, they will be around the matrix
    t = obtain_rectangular_submatrices(m)[0]  # Take just the first tuple: U=t[0], D=t[1], L=t[2], R=t[3]
    t = (t[0]+1, t[1]+1, t[2]+1, t[3]+1)  # The previous calculation is done using Numpy, so it is Zero based. Correct this

    # Obtain the parent, child and factor being scaled
    try:
        parent_child = sh.cell(row=t[0] - 1, column=t[2] - 1).value
        child, parent = parent_child.split("/")
        parent_processor_type = parent.strip()
        child_processor_type = child.strip()
    except:
        issues.append((3, "Could not obtain the parent and child processor types: '"+parent_child+"' in upscale command"))
        some_error = True

    scaled_factor = sh.cell(row=t[0]-2, column=t[2]-1).value
    if not scaled_factor:
        issues.append((3, "Factor not specified in upscale command"))
        some_error = True

    if some_error:
        return issues, None, None

    # # Merged cells
    # for ra in sh.merged_cell_ranges:
    #     t = openpyxl.utils.range_boundaries(ra)  # min col, min row, max col, max row (max's included)
    #     mc = (t[1]-1, t[3]-1, t[0]-1, t[2]-1)  # Rearrange and subtract one
    #     v = m[mc[0], mc[2]]
    #     m[mc[0]:mc[1]+1, mc[2]:mc[3]+1] = v

    # Analyze rows above the matrix, to see code lists
    subrows = []
    for r in range(t[0]-1, 0, -1):
        # Get the whole subrow into a list
        subrow = get_subrow(r, t[2], t[3])
        # If empty, break
        if not any(subrow):
            break
        subrows.append(subrow)

    # Analyze cols to the left of the matrix, to see code lists
    subcols = []
    for c in range(t[2]-1, 0, -1):
        # Get the whole subcolumn into a list
        subcol = get_subcolumn(c, t[0], t[1])
        # If empty, break
        if not any(subcol):
            break
        subcols.append(subcol)

    # "scales" list will contain a list of dictionaries with the keys "codes" which are the codes leading to a cell
    # in the table, and "weight" the number in the cell. To form "codes" a prefix, formed by codes in the columns to the
    # left, is concatenated to a suffix, formed by the codes in the rows on top of the matrix

    # Obtain suffix tuples from columns (as lists really)
    sufix_tuples = []
    for i, c in enumerate(range(t[2], t[3])):
        sufix_tuples.append([subrows[j][i] for j in range(len(subrows))])

    # Iterate every row, col in the factors matrix to obtain the SCALES list
    scales = []
    for i, r in enumerate(range(t[0], t[1])):
        prefix_tuple = [subcols[j][i] for j in range(len(subcols))]  # Subtuple from subcols
        for j, c in enumerate(range(t[2], t[3])):
            w = sh.cell(row=r, column=c).value
            scales.append(dict(codes=prefix_tuple+sufix_tuples[j], weight=w))

    # # Parent Processor type	Child Processor type	Scaled Factor	Source
    # parent_processor_type = None
    # child_processor_type = None
    # scaled_factor = None
    # source = None
    # for c in range(area[2], area[3]):  # Columns
    #     key = sh.cell(row=1, column=c).value
    #     value = sh.cell(row=2, column=c).value
    #     if key.lower() in ["parent"]:
    #         parent_processor_type = value
    #     elif key.lower() in ["child"]:
    #         child_processor_type = value
    #     elif key.lower() in ["scaled factor"]:
    #         scaled_factor = value
    #     elif key.lower() in ["source"]:  # "Observer"
    #         source = value

    content = {"parent_processor_type": parent_processor_type,
               "child_processor_type": child_processor_type,
               "scaled_factor": scaled_factor,
               "source": None,
               "scales": scales
               }
    label = "Upscale child '"+child_processor_type+"' into parent '"+parent_processor_type+"'"
    return issues, label, content


